import jax.numpy as jnp


def com_1D(positions, masses=None, axis=0) -> float:
    """Center of mass in a single axis.

    Args:
        positions (array): Array with 3D positions.
        masses (array, optional): Mass at each position point. If None, all masses are unitary. Defaults to None.
        axis (int, optional): Axis along which to compute the center of mass of. Defaults to 0 (first).

    Raises:
        Exception: Data is not a 2D array.

    Returns:
        float: Center of mass.
    """

    if masses is None:
        masses = jnp.ones((len(positions)))

    if positions.ndim == 2:
        xs = positions[:, axis]
    else:
        raise Exception("Only implemented for 2d data")

    return jnp.sum(xs * masses) / len(xs)


def vel_disp_1D(velocities, axis=0) -> float:
    """Velocity dispersion

    Args:
        velocities (array): Array with 3D velocities and positions.
        axis (int, optional): Axis along which to compute the center of mass of. Defaults to 0 (first).

    Raises:
        Exception: Data is not a 3D array.

    Returns:
        float: Center of mass.
    """
    if velocities.ndim == 3:
        vxs = velocities[:, 1, axis]
    else:
        raise Exception("Only implemented for 3d data")

    return jnp.std(vxs)


if __name__ == "__main__":
    pass



from jaxknife.jaxknife import bootstrap_resamples
from make_testdata import *


def test_bootstrap_resamples():

    # Usual Test Case
    n_resamples = 100
    size = 100
    seed = 0
    testdata1D = make_pos_data(size=size,seed=seed)

    resamples = bootstrap_resamples(testdata1D,n_resamples=n_resamples,seed=seed)
    
    assert resamples.shape == (n_resamples,size,3), "bootstrap_resamples() returned the wrong shape in 1D data case."
    assert resamples[0][0] in testdata1D, "bootstrap_resamples() invented new data in 1D data case."

    testdata2D = make_posvel_data(size=size,seed=seed)

    resamples = bootstrap_resamples(testdata2D,n_resamples=n_resamples,seed=seed)
    
    assert resamples.shape == (n_resamples,size,2,3), "bootstrap_resamples() returned the wrong shap in ND data case."
    assert resamples[0][0] in testdata2D, "bootstrap_resamples() invented new data in ND data case."


from jaxknife.jaxknife import bootstrap
from make_testdata import *
from jaxknife.stats_funcs import com_1D
import pytest

def test_bootstrap():
    
    n_resamples = 1000
    size = 100
    seed = 0

    pos_data = make_pos_data(size=size,seed=seed)
    com_arr = bootstrap(pos_data,com_1D,n_resamples=n_resamples,seed=seed)
    
    assert np.shape(com_arr) == (n_resamples,), 'bootstrap() returned the wrong shape of stats array.'
    assert np.median(com_arr) == pytest.approx(0,abs=0.1), 'bootstrap() resulted in a weird center of mass.'


import jax
import jax.numpy as jnp


def bootstrap(data, statistic_func, n_resamples, seed):
    """Bootstrap.

    Args:
        data (array): Input data with shape (N, ...) where N is the number of realizations of the observable.
        statistic_func (callable): Function to calculate a statistic on one set of realizations of the observable.
        n_resamples (int): Number of times to resample the data.
        seed (int): Seed for RNG.

    Returns:
        jax array: The value of the statistic for each resample.
    """
    resamples = bootstrap_resamples(data, n_resamples, seed)
    est_arr = _apply_vect_statistic(statistic_func, resamples)
    return est_arr


def jackknife(data, statistic_func):
    """Jackknife.

    Args:
        data (array): Input data with shape (N, ...) where N is the number of realizations of the observable.
        statistic_func (callable): Function to calculate a statistic on one set of realizations of the observable.

    Returns:
        jax array: The value of the statistic for each resample.
    """
    resamples = jackknife_resamples(data)
    est_arr = _apply_vect_statistic(statistic_func, resamples)
    return est_arr


def _apply_vect_statistic(statistic_func, resamples):
    """Vectorize and apply statistic function.

    Args:
        statistic_func (callable): Function to calculate a statistic on one set of realizations of the observable.
        resamples (jax array): Resampled data with shape (n_resamples, N, ...)

    Returns:
        jax array: The value of the statistic for each resample.
    """
    statistic_vectorized = jax.vmap(statistic_func)
    return statistic_vectorized(resamples)


def bootstrap_resamples(data, n_resamples, seed):
    """Generate bootstrap resamples.

    Args:
        data (array): Input data with shape (N, ...) where N is the number of realizations of the observable.
        n_resamples (int): Number of times to resample the data.
        seed (int): Seed for RNG.

    Returns:
        jax array: Resampled data with shape (n_resamples, N, ...)
    """
    key = jax.random.PRNGKey(seed)
    resamples_size = (n_resamples, data.shape[0])
    resamples = jax.random.choice(key, data, resamples_size, replace=True)
    return resamples


def jackknife_resamples(data):
    """Generate jackknife resamples.

    Args:
        data (array): Input data with shape (N, ...) where N is the number of realizations of the observable.

    Returns:
        jax array: Resampled data with shape (N, N, ...)
    """
    redata = jnp.repeat(jnp.array([data]), data.shape[0], axis=0)
    mask = jnp.repeat(
        ~jnp.eye(data.shape[0], dtype=bool).reshape(data.shape[0], data.shape[0], -1),
        data.shape[-1],
        axis=2,
    )
    return redata[mask].reshape(data.shape[0], data.shape[0] - 1, -1)


if __name__ == "__main__":
    pass
